#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <unistd.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "lwm.h"

Client *current;
static Client *clients;

static int popup_width;	/* The width of the size-feedback window. */

Edge interacting_edge;

static void sendClientMessage(Window, Atom, long, long);

void
setactive(Client *c, int on, long timestamp) {
	int inhibit = isShaped(c->window);

	if (c == 0 || hidden(c))
		return;

	if (!inhibit) {
		XMoveResizeWindow(dpy, c->parent,
			c->size.x, c->size.y - titleHeight(),
			c->size.width, c->size.height + titleHeight());
		XMoveWindow(dpy, c->window, border, border + titleHeight());
	}

	if (on && c->accepts_focus) {
		XSetInputFocus(dpy, c->window, RevertToPointerRoot, CurrentTime);
		if (c->proto & Ptakefocus)
			sendClientMessage(c->window, wm_protocols, wm_take_focus,
				timestamp);
		cmapfocus(c);
	}

	if (!inhibit)
		Client_DrawBorder(c, on);
}


void Client_DrawName(Client *c)
{
	int i;
		
	/* Draw window title. */
	if (c != NULL && c->name != 0) {
		
		i=strlen(c->name);
		
		/* Find a shortened name to fit */
		/* FIXME: add '..' ?? */
		while(XTextWidth(font, c->name, i) > 120)
			i--;
			
		XFillRectangle(dpy, c->screen->root, c->screen->size_gc,
			4, 0,
			120,
			font->ascent + font->descent);
		XDrawString(dpy, c->screen->root, c->screen->gc, 4,
			font->ascent + font->descent,
			c->name, i);
	}
	else
	{
		XFillRectangle(dpy, c->screen->root, c->screen->size_gc,
			4, 0,
			120,
			font->ascent + font->descent);
	}
}

void
Client_DrawBorder(Client *c, int active) {
	int quarter = (border + titleHeight()) / 4;

	if (c == NULL || c->parent == c->screen->root || c->parent == 0)
		return;

	XSetWindowBackground(dpy, c->parent,
		active ? c->screen->black : c->screen->gray);
	XClearWindow(dpy, c->parent);

	/* Draw the ``box''. */
	if (active) {
		XDrawRectangle(dpy, c->parent, c->screen->gc,
			quarter + 2, quarter, 2 * quarter, 2 * quarter);
	}
}


Client *
Client_Get(Window w) {
	Client * c;
	
	if (w == 0 || (getScreenFromRoot(w) != 0))
		return 0;
	
	/* Search for the client corresponding to this window. */
	for (c = clients; c; c = c->next)
		if (c->window == w || c->parent == w)
			return c;
	
	/* Not found. */
	return 0;
}


Client *
Client_Add(Window w, Window root) {
	Client * c;

	if (w == 0 || w == root)
		return 0;

	/* Search for the client corresponding to this window. */
	for (c = clients; c != 0; c = c->next)
		if (c->window == w || c->parent == w)
			return c;

	c = calloc(1, sizeof *c);
	c->window = w;
	c->parent = root;
	c->hidden = False;
	c->state = WithdrawnState;
	c->internal_state = INormal;
	c->cmap = None;
	c->name = c->fallback_name = 0;
	c->ncmapwins = 0;
	c->cmapwins = 0;
	c->wmcmaps = 0;
	c->accepts_focus = 1;
	c->next = clients;

	/* Add to head of list of clients. */
	return (clients = c);
}


void
Client_Remove(Client *c) {
	Client * cc;
	char *name;
	
	if (c == 0)
		return;

	if (c->name != 0)
	{
		XFree(c->name);
		c->name = NULL;
	}
	Client_DrawName(c);	/* Clean up */
	
	/* Remove the client from our client list. */
	if (c == clients)
		clients = c->next;
	for (cc = clients; cc && cc->next; cc = cc->next) {
		if (cc->next == c)
			cc->next = cc->next->next;
	}
	
	/* A deleted window can no longer be the current window. */
	if (c == current) {
		current = 0;
	}
	
	
	if (getScreenFromRoot(c->parent) == 0)
		XDestroyWindow(dpy, c->parent);
	
	if (c->ncmapwins != 0) {
		XFree(c->cmapwins);
		free(c->wmcmaps);
	}
	/*
	 *	New top application ?
	 */
	c=Client_TopApp();
	if(c)
		Client_DrawName(c);
}


void
Client_Back(Client *c)
{
	Client **cp= &clients;
	if (c == 0)
		return;

	XLowerWindow(dpy, c->window);
	XLowerWindow(dpy, c->parent);
	
	/*
	 *	Amend client list as we are now bottom
	 */
	
	while((*cp)!=NULL)
	{
		if((*cp)==c)
		{
			*cp=c->next;
			continue;
		}
		cp = &((*cp)->next);
	}	
	*cp = c;
	c->next = NULL;
	Client_DrawName(Client_TopApp());
}

void
Client_Top(Client *c)
{
	Client **cp= &clients;
	if (c == 0)
		return;

	/*
	 *	Amend client list as we are now top
	 */
	
	while((*cp)!=NULL)
	{
		if((*cp)==c)
		{
			*cp=c->next;
			break;
		}
		cp = &((*cp)->next);
	}	
	c->next = clients;
	clients = c;
	Client_DrawName(Client_TopApp());
}

void
Client_Close(Client *c) {
	if (c == 0)
		return;

	/*
	 *	Terminate the client nicely if possible. Be brutal otherwise.
	 */
	if (c->proto & Pdelete) {
		sendClientMessage(c->window, wm_protocols, wm_delete, CurrentTime);
	} else {
		XKillClient(dpy, c->window);
	}
}

Client *Client_TopApp(void)
{
	Client *c=clients;
	while(c && !c->appwin)
		c=c->next;
	return c;
}

void
Client_SetState(Client *c, int state) {
	long	data[2];

	data[0] = (long) state;
	data[1] = (long) None;

	c->state = state;
	XChangeProperty(dpy, c->window, wm_state, wm_state, 32,
		PropModeReplace, (unsigned char *) data, 2);
}

static void
sendClientMessage(Window w, Atom a, long data0, long data1) {
	XEvent	ev;
	long	mask;

	memset(&ev, 0, sizeof(ev));
	ev.xclient.type = ClientMessage;
	ev.xclient.window = w;
	ev.xclient.message_type = a;
	ev.xclient.format = 32;
	ev.xclient.data.l[0] = data0;
	ev.xclient.data.l[1] = data1;
	mask = (getScreenFromRoot(w) != 0) ? SubstructureRedirectMask : 0L;

	XSendEvent(dpy, w, False, mask, &ev);
}

extern void
Client_FreeAll(void) {
	Client *c;
	XWindowChanges wc;

	for (c = clients; c; c = c->next) {
		int not_mapped = !normal(c);

		/* Remap the window if it's hidden. */
		if (not_mapped) {
			XMapWindow(dpy, c->parent);
			XMapWindow(dpy, c->window);
		}

		/* Reparent it, and then push it to the bottom if it was hidden. */
		XReparentWindow(dpy, c->window, c->screen->root, c->size.x, c->size.y);
		if (not_mapped)
			XLowerWindow(dpy, c->window);

		/* Give it back its initial border width. */
		wc.border_width = c->border;
		XConfigureWindow(dpy, c->window, CWBorderWidth, &wc);
	}
}

extern void
Client_ColourMap(XEvent *e) {
	int i;
	Client * c;

	for (c = clients; c; c = c->next) {
		for (i = 0; i < c->ncmapwins; i++) {
			if (c->cmapwins[i] == e->xcolormap.window) {
				c->wmcmaps[i] = e->xcolormap.colormap;
				if (c == current)
					cmapfocus(c);
				return;
			}
		}
	}
}

extern int
hidden(Client *c) {
	return c->state == IconicState;
}

extern int
withdrawn(Client *c) {
	return c->state == WithdrawnState;
}

extern int
normal(Client *c) {
	return c->state == NormalState;
}
